import r3make.utils as utils

from rich.table import Table
from rich.panel import Panel
from rich.console import Console

from r3make.commands import CBUILD_COMMANDS
from r3make.version import YEAR, MINOR, PATCH

from r3make.compiler import SUPPORTED_COMPILERS, BaseCompiler
from r3make.fetch import fetch_files, fetch_library, fetch_compiler, fetch_compiler_instance

def _build_summary(
        console:Console,
        c_instance,
        c_flags,
        c_defines,
        src_dirs,
        src_files,
        inc_dirs,
        lib_links,
        out_dir,
        out_type,
        out_name
        ) -> None:
    
    ver = f"{YEAR}.{MINOR}.{PATCH}"
    table = Table(title=f"Build Summary: {out_name}", caption=f"[bold green]r3make[/] [bold white]{ver}[/]", show_lines=True)
    table.add_column("Field", style="bold green", no_wrap=True)
    table.add_column("Value", style="bold cyan")
    table.add_row("Compiler", c_instance.name)
    table.add_row("Flags", " ".join(c_flags))
    table.add_row("Defines", " ".join(c_defines))
    table.add_row("Source Files", str(len(src_files)))
    table.add_row("Source Dirs", " ".join(src_dirs))
    table.add_row("Include Dirs", " ".join(inc_dirs))
    table.add_row("Libraries", " ".join(list(lib_links.keys())))
    table.add_row("Output Dir", out_dir)
    table.add_row("Output Type", out_type)
    
    console.print(table)

def r3make_build(console:Console, config) -> None:
    if not isinstance(config, dict): return None

    # extract compiler configuration
    if not fetch_compiler(config["c-instance"]):
        console.print(Panel(f"[bold red]r3make[/] Compiler not found on system: {config["c-instance"]}", expand=False))
        return None

    c_instance:BaseCompiler = fetch_compiler_instance(config["c-instance"])
    if not isinstance(c_instance, BaseCompiler):
        console.print(Panel(f"[bold red]r3make[/] Compiler not supported: {config["c-instance"]}", expand=False))
        return None
    
    c_flags = []
    c_defines = []
    if len(config["c-flags"]): 
        c_flags = [ c_flag.strip() for c_flag in config["c-flags"] ]
    if len(config["c-defines"]): 
        c_defines = [ c_define.strip() for c_define in config["c-defines"] ]

    # extract source file configuration
    src_dirs = []
    for src_dir in config["src-dirs"]:
        if not utils.os.path.exists(src_dir):
            print(f"Source Directory Not Found: {src_dir}")
            continue
        src_dirs.append(src_dir)
    
    src_files = []
    for src_dir in src_dirs:
        for src_file in fetch_files(".c", src_dir):
            src_files.append(src_file.strip())
    
    # extract include file configuration
    inc_dirs = []
    for inc_dir in config["inc-dirs"]:
        if not utils.os.path.exists(inc_dir):
            print(f"Include Directory Not Found: {inc_dir}")
            continue
        inc_dirs.append(inc_dir)
    
    # extract linked library configuration
    lib_links = {}
    if len(config["lib-links"]):
        for lib, path in config["lib-links"].items():
            if path == None:
                flib, fpath = fetch_library(lib, path)
                if fpath == "OS":
                    lib_links[flib] = None
                    continue
                elif fpath == None or not utils.os.path.exists(fpath):
                    print(f"Library Not Found: {flib}")
                    return
                lib_links[flib] = fpath
            elif utils.os.path.exists(path):
                for ext in ['.dll', '.a', '.lib', '.so']:
                    if f"{path}{utils.SEP}{lib}{ext}" in fetch_files(ext, path):
                        lib_links[lib] = path
                        break
                    else:
                        print(f"Library Not Found: {lib}")
                        return

    # extract output configuration
    out_dir = config["out-dir"]
    if not utils.os.path.exists(out_dir):
        utils.os_mkdir(out_dir)
    
    out_type = config["out-type"]
    if out_type not in c_instance.out_types:
        print(f"({c_instance.name}) Invalid Output Type: {out_type}")

    out_name = config["out-name"]

    # extract r3make commands
    if config.get("r3make", None):
        pre_commands = config["r3make"]["pre-build"]
        for command in pre_commands:
            if command in CBUILD_COMMANDS:
                CBUILD_COMMANDS[command](config, param=pre_commands[command])

    _build_summary(
        console,
        c_instance,
        c_flags,
        c_defines,
        src_dirs,
        src_files,
        inc_dirs,
        lib_links,
        out_dir,
        out_type,
        out_name
    )

    # compile source code into object files
    if not utils.os.path.exists(f"{config["out-dir"]}{utils.SEP}ofiles"):
        try:
            utils.os.mkdir(f"{config["out-dir"]}{utils.SEP}ofiles")
        except Exception as e:
            print(e)
            return None

    ofiles:list[str] = c_instance.compile(
        c_flags,
        c_defines,
        src_files,
        inc_dirs,
        out_dir
    )

    # link object files into final output
    result:bool = c_instance.link(
        ofiles,
        lib_links,
        out_name,
        out_type,
        out_dir
    )

    if config.get("r3make", None):
        post_commands = config["r3make"]["post-build"]
        for command in post_commands:
            if command in CBUILD_COMMANDS:
                CBUILD_COMMANDS[command](config, param=post_commands[command])

    if isinstance(result, bool) and result == False:
        console.print(Panel(f"[bold red]r3make[/] Build Failed", expand=False))
    elif isinstance(result, bool) and result == True:
        console.print(Panel(f"[bold green]r3make[/] Build Success", expand=False))



import r3make.utils as utils
from r3make.compiler import SUPPORTED_COMPILERS, BaseCompiler

DEFAULT_SEARCH_PATHS: dict[str, list[str]] = {}

def fetch_compiler(compiler_name) -> bool:
    """ Tests to see if a compiler exists/is callable by outputting version info by running: `{compiler.prefix} -v`"""
    try:
        result = utils.subprocess.run(
            f"{compiler_name.lower()} -v",
            text=True,
            check=True,
            capture_output=True,
            shell=True
        )
        output = result.stderr
        return True
    except utils.subprocess.CalledProcessError:
        return False

def define_search_paths() -> None:
    """
    Defines the default search paths for all supported compilers and platforms,
    including both archives and binaries.
    """

    # common os-specific default paths
    os_paths = {
        "Windows": [
            utils.os_path("C:\\Windows\\System32"),         # this causes linker errors when used as the -L directory for some reason?
            utils.os_path("C:\\Windows\\System32\\wbem")    # so based on the above notice, windows users should put their installed/custom libs here
        ],
        "Linux": [
            utils.os_path("/lib"),
            utils.os_path("/usr/lib"),
            utils.os_path("/usr/local/lib")
        ],
        "Darwin": [
            utils.os_path("/lib"),
            utils.os_path("/usr/lib"),
            utils.os_path("/usr/local/lib"),
            utils.os_path("/opt/homebrew/lib")  # Homebrew on ARM macutils.OS
        ]
    }

    # os-specific defaults
    global DEFAULT_SEARCH_PATHS
    DEFAULT_SEARCH_PATHS["OS"] = os_paths.get(utils.platform.system(), [])

    # get paths from environment variables
    env_paths = {
        "Windows": ["LIB"],  # MSVC uses LIB for library paths
        "Linux": ["LD_LIBRARY_PATH"],
        "Darwin": ["DYLD_LIBRARY_PATH"]
    }
    
    for env_var in env_paths.get(utils.platform.system(), []):
        paths = utils.os.environ.get(env_var, "").split(utils.os.pathsep)
        DEFAULT_SEARCH_PATHS["ENV"] = [utils.os_path(p) for p in paths if p]

    # get the defualt search dirs from each supported compiler
    for name, compiler in SUPPORTED_COMPILERS.items():
        if not fetch_compiler(name): continue

        DEFAULT_SEARCH_PATHS[name] = []
        search_command = f"{compiler.prefix} -print-search-dirs"

        try:
            result = utils.subprocess.run(
                search_command,
                text=True,
                check=True,
                capture_output=True,
                shell=True
            )
            output = result.stdout
        except utils.subprocess.CalledProcessError:
            # fallback to utils.os.popen if utils.subprocess fails (python and environment variables arent cool)
            try:
                with utils.os.popen(search_command) as proc:
                    output = proc.read()
                continue
            except FileNotFoundError:
                print(f"Unable to call compiler: {name}")
                continue

        for line in output.splitlines():
            if line.startswith("libraries: =") or line.startswith("programs: ="):
                paths = line.split("=", 1)[1].strip().split(utils.os.pathsep)
                DEFAULT_SEARCH_PATHS[name].extend(paths)

def fetch_library(lib:str, path: str) -> list[str]:
    """
    Recursively searches for library files (.dll, .a, .lib, .so) in the specified directory.
    Returns True if any library is found, otherwise False.
    """
    if isinstance(path, str) and utils.os.path.exists(path):
        current_dirs = [utils.os_path(path), utils.os.getcwd()]
    else:
        current_dirs = [utils.os.getcwd()]

    for search_path in DEFAULT_SEARCH_PATHS.values():
        current_dirs.extend(search_path)

    # Search recursively across all applicable directories
    for directory in set(current_dirs):
        for root, _, files in utils.os.walk(directory.strip()):
            for file in files:
                if file.endswith(('.dll', '.a', '.lib', '.so')) and file.split(".")[0] == lib:
                    if root in DEFAULT_SEARCH_PATHS["OS"]:  # OS default libs dont need a link path
                        return [lib, "OS"]
                    return [lib, root]
    return [lib, None]  # lib not found on the system

def fetch_compiler_instance(compiler_name) -> BaseCompiler:
    try:
        return SUPPORTED_COMPILERS[compiler_name]
    except (KeyError) as e:
        print(f"Unsupported compiler: {compiler_name}")
        return None

def fetch_files(extension:str, directory:str) -> list[str]:
    directory = utils.os_path(directory)
    source_files = []
    if not utils.os.path.exists(directory): return source_files
    for root, _, files in utils.os.walk(directory.strip()):
        for file in files:
            if file.endswith(extension):
                source_files.append(utils.os.path.join(root, file))
    return source_files
